# -*- coding: utf-8 -*-
from importlib import import_module
import json
import os
import os.path
import random

from .exceptions import ConfigError, SourceError

SEP = ':'
WILDCARD = '*'


def get_rnd_item(fpath):
    """Get random item from list read from JSON data.

    :param fpath: the filepath of the JSON containing a list
    :type fpath: str
    :returns: str - a user agent string chosen randomly from list
    """
    try:
        with open(fpath, 'r') as json_file:
            items = json.load(json_file)
    except ValueError as e:
        raise ConfigError(e)
    return random.choice(items)


def get_uuid(*args):
    """Return a UUID built from fragments

    :param \*args: UUID fragments
    :type \*args: str
    :returns: str - a UUID resulting from string concatenation
    """
    return SEP.join([frag.replace(SEP + WILDCARD, "") for frag in args])


def passthrough_loader(base_uuid, data):
    return {base_uuid: data}


def dig_loader(base_uuid, data):
    _data = {}
    for key, info in data.items():
        name = get_uuid(base_uuid, key)
        _data[name] = info
    return _data


def get_source(name):
    """Get the source object based on a name."""
    # import only occurs when function is called
    from .conf import settings
    if name not in settings['SOURCES']:
        raise SourceError("Unknown source ({})".format(name))
    modname = settings['SOURCES'][name]['modname']
    mod = import_module(modname)
    src = mod.Source(name)
    return src


def load_backend(backend_name):
    """Load a backend based on a module name

    :param backend_name: database backend to use
    :type backend_name: str
    :returns: module
    """
    try:
        return import_module(backend_name)
    except ImportError as e:
        # FIXME add error treatment
        raise


def load_tree(base_dir, data_loader):
    """Create a one-dimensional dictionary given a tree directory structure.

    The keys are generated with a separator per folder depth.

    .. note::

        Files in the filesystem are expected to have a ``.json`` suffix.

    :param base_dir: a valid directory or path
    :type base_dir: str
    :param data_loader: a function specificying how to name the keys
    :type data_loader: func
    :returns: dict
    """
    tree = {}
    for root, dirs, files in os.walk(base_dir):
        relpath = os.path.relpath(root, base_dir)
        if relpath != os.curdir:
            reluuid = relpath.replace(os.sep, SEP)
        else:
            reluuid = ''

        for file in files:
            if not file.endswith('.json'):
                continue

            filepath = os.path.join(root, file)
            try:
                with open(filepath, 'r') as jsonfd:
                    data_chunk = json.load(jsonfd)
            except ValueError as e:
                # "No JSON object could be decoded"
                msg = "{} ({})".format(e, filepath)
                raise ConfigError(msg)
            except IOError as e:
                raise ConfigError("{} ({})"
                                  .format(e.strerror, filepath))
            src_name = os.path.splitext(file)[0]
            if reluuid == '':
                uuid_key = src_name
            else:
                uuid_key = SEP.join((reluuid, src_name))
            tree.update(data_loader(uuid_key, data_chunk))

    return tree



# -*- coding: utf-8 -*-
"""
Naive settings for openkongqi

Very naive approach for settings management. 

No lazy loading, all straight talk!
"""

from __future__ import absolute_import, print_function, unicode_literals
from importlib import import_module
import logging.config
import os
import sys

from .exceptions import ConfigError, SourceError
from .utils import load_tree, dig_loader
from .status.base import create_statusdb
from .records.base import create_recsdb
from .cache.base import create_cachedb


here = os.path.abspath(os.path.dirname(__file__))

#CONFMODULE = 'okqconfig'
#os.environ.setdefault('OKQ_CONFMODULE', CONFMODULE)

# default settings
global_settings = {
    'RESOURCE_CACHE': '_cache',
    'DATABASES': {
        'status': {
            'ENGINE': 'openkongqi.status.redisdb',
            'HOST': 'localhost',
            'PORT': 6379,
            'DB_ID': 1,
        },
        'records': {
            'ENGINE': 'openkongqi.records.sqlite3',
            'NAME': 'openkongqi',
        },
        'cache': {
            'ENGINE': 'openkongqi.cache.redisdb',
            'HOST': 'localhost',
            'PORT': 6379,
            'DB_ID': 0,
        },
    },
    'DEBUG': False,
    'SOURCES': {},
    'SOURCES_DIR': os.path.join(here, 'data/sources'),
    'FEEDS': {},
    'UA_FILE': os.path.join(here, 'data/user_agent_strings.json'),
    'STATIONS_MAP_DIR': os.path.join(here, 'data/stations'),
    'API_KEYS': {}
}

settings = {}
statusdb = None
cachedb = None
recsdb = None
file_cache = None


def config_from_object(obj):
    global statusdb, cachedb, recsdb, file_cache
    local_settings = obj.settings
    # use default values or the values found in settings file
    for setting in global_settings:
        settings[setting] = local_settings.get(setting,
                global_settings[setting])

    # check for existence and readability of UA_FILE;
    # catch ConfigError as early as possible
    try:
        with open(settings['UA_FILE'], 'r'):
            pass
    except IOError as e:
        raise ConfigError("{} ({})"
                          .format(e.strerror, settings['UA_FILE']))

    # check for existence of sources and stations map directory
    for fname in ['STATIONS_MAP_DIR', 'SOURCES_DIR']:
        fpath = settings[fname]
        if not os.path.exists(fpath):
            raise ConfigError("Directory not found ({}: {})"
                              .format(fname, fpath))

    # load sources from sources directory
    settings['SOURCES'] = load_tree(settings['SOURCES_DIR'], dig_loader)
    if not settings['SOURCES']:
        raise SourceError("No configured source")

    # load status db
    statusdb = create_statusdb(settings['DATABASES']['status'])

    # load cache db
    cachedb = create_cachedb(settings['DATABASES']['cache'])

    # load records db
    recsdb = create_recsdb(settings['DATABASES']['records'], cachedb)

    # create instance of cache and catch any error as early as possible
    from .filecache import FileCache
    file_cache = FileCache(settings['RESOURCE_CACHE'])



# -*- coding: utf-8 -*-

from ..utils import load_backend


class BaseStatusWrapper(object):
    """Base wrapper class to get database status. This class is to be used
    as parent of any database status wrapper class.

    The children wrapper classes will have to define custom methods for
    creating the database connection.
    """

    def __init__(self, db_settings, *args, **kwargs):
        self._cnx = self.create_cnx(db_settings)

    def create_cnx(self, db_settings):
        """Create a connection to the database.

        .. warning:: This method has to be overwritten

        :param db_settings: dictionary containing database configuration
        :type db_settings: dict
        """
        raise NotImplementedError

    def db_init(self):
        """Initialize database.

        .. warning:: This method has be overwritten
        """
        raise NotImplementedError

    def set_status(self, name, data):
        """Save the status.

        .. warning:: This method has to be overwritten
        """
        raise NotImplementedError

    def get_status(self, name):
        """Return the status with the given name and arguments.

        .. warning:: This method has to be overwritten

        :param name: the name as found in `settings.SOURCES`
        :type name: str
        """
        raise NotImplementedError


def create_statusdb(db_settings):
    """Create a status database instance given a database settings dict."""
    status_db_mod = load_backend(db_settings['ENGINE'])
    return status_db_mod.StatusWrapper(db_settings)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from datetime import datetime
import os
import shutil

from .exceptions import CacheError


FILENAME = "{key}-{ts}.txt"


class FileCache(object):
    """Simple file caching based on a key name and content.
    """
    def __init__(self, folder_name):
        if os.path.isabs(folder_name):
            self.cachepath = folder_name
        else:
            self.cachepath = os.path.join(
                os.getcwd(),
                folder_name)
        if not os.path.exists(self.cachepath):
            try:
                os.makedirs(self.cachepath)
            except OSError as e:
                raise CacheError(
                    'cache creation problem ({})'.format(e.strerror))

    def get_fp(self, key, ts=None):
        # TODO doc
        if ts is None:
            ts = datetime.now()

        return os.path.join(self.cachepath,
                            FILENAME.format(key=key,
                                            ts=ts.strftime('%Y%m%d%H%M%S')))

    def get_latest_fp(self, key):
        # TODO doc
        return os.path.join(self.cachepath,
                            FILENAME.format(key=key, ts='latest'))

    def set(self, key, fsrc, ts=None):
        # TODO doc
        filename = self.get_fp(key, ts)
        if not os.path.exists(os.path.dirname(filename)):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as e:
                raise CacheError(
                    'cache creation problem ({})'.format(e.strerror))
        fdst = open(filename, 'wb')
        shutil.copyfileobj(fsrc, fdst)
        filelink = self.get_latest_fp(key)
        if os.path.lexists(filelink):
            os.remove(filelink)
        os.symlink(os.path.basename(filename), filelink)

    def get(self, key, ts=None):
        # TODO doc
        try:
            fd = open(self.get_fp(key, ts), 'r')
        except IOError:
            return
        return fd

    def get_latest(self, key):
        try:
            fd = open(self.get_latest_fp(key), 'r')
        except IOError:
            return
        return fd



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from datetime import datetime
import io
import logging
import os
import re
from urllib3.util.retry import Retry
from urllib3.exceptions import MaxRetryError

from celery.utils.log import get_task_logger
import requests

from ..apikeys import get_api_key
from ..conf import settings, statusdb, recsdb, file_cache
from ..exceptions import SourceError
from ..stations import get_station_map
from ..utils import get_rnd_item, get_uuid

import pytz


logger = get_task_logger(__name__)


class BaseSource(object):
    """Base source class to scrape online resources. This class is to be used
    as parent of any scraping class.

    Basic mechanism from fetching online resource to hosing extracted data in
    the database, :meth:`openkongqi.source.BaseSource.scrape` is the main entry
    point of this class.
    """

    key_context = None
    _now = None

    def __init__(self, name):
        """
        :param name: fetch name
        :type name: str
        """
        if name not in settings['SOURCES']:
            raise SourceError('Unknown source ({})'.format(name))
        self.name = name
        self.target = settings['SOURCES'][name]['target']
        self._station_map = get_station_map(
            settings['SOURCES'][name]['uuid'])
        self._api_key = get_api_key(name.split(":")[0])
        self._tz = pytz.timezone(settings['SOURCES'][name]['tz'])
        self._status = statusdb
        self._cache = file_cache
        self._records = recsdb

    def scrape(self):
        """Main entry point for :class:`openkongqi.source.BaseSource` instances.

        Orchestrates the resource scraping, following actions are performed:

        * :meth:`openkongqi.source.BaseSource.fetch`: fetch online resource
        * :meth:`openkongqi.source.BaseSource.save_status`: save fetching
          status
        * :meth:`openkongqi.source.BaseSource.cache`: save the resource in the
          cache
        * :meth:`openkongqi.source.BaseSource.extract`: extract the data
          from the resource
        * :meth:`openkongqi.source.BaseSource.save_data`: save extracted data
        """
        self._now = datetime.now(pytz.utc)
        src_content = self.fetch()
        self.save_status()
        if src_content is not None:
            content = self.cache(src_content)
            data = self.extract(content)
            self.save_data(data)

    def fetch(self):
        """Fetch the resource

        .. warning:: This method has to be overwritten

        :returns: content - a file-like object
        """
        raise NotImplementedError

    def post_fetch(self, resource):
        """Treat the resource after fetching raw content

        .. warning:: This method has to be overwritten

        :param resource: raw content / data (could be http, csv, xml, etc...)
        :returns: content - a file-like object
        """
        raise NotImplementedError

    def get_station_uuid(self, name):
        """Return the uuid of a station

        .. note:: This is not performing any recursive search in the map
        """
        return get_uuid(settings['SOURCES'][self.name]['uuid'],
                        self._station_map[name]['uuid'])

    def get_status_data(self):
        """Get the fetch status

        :returns: data - a dict with data to serialize in the status entry
        """
        return dict()

    def save_status(self):
        """Save status data to keep track of fetching history.

        Uses :meth:`openkongqi.source.BaseSource.get_status_data` to get the
        status data to save.
        """
        data = self.get_status_data()
        if data is None:
            data = {'ts': self._now.strftime('%Y%m%d%H%M%S')}
        else:
            data.setdefault('ts', self._now.strftime('%Y%m%d%H%M%S'))
        self._status.set_status(self.name, data)

    def cache(self, content):
        """Cache fetching content

        :param content: content to cache
        :type content: file-like object
        """
        self._cache.set(self.name, content, self._now)
        fp = self._cache.get_fp(self.name, self._now)
        # display how much is cached into server
        self.log_info("Cached {} kilobytes to server."
                    .format(os.path.getsize(fp)))
        return open(fp, 'rb')

    def pythonify(self, text, is_num=False):
        if text is None:
            return None

        if self.null_re is not None:
            if self.null_re.match(text) is None:
                if is_num:  # remove everything but nums and dots
                    try:
                        return float(re.sub(r'[^\d.]+', '', text))
                    except ValueError:
                        self.log_error(f"{self.name} - Could not convert value: {text}")
                else:
                    return text
            else:
                return None
        else:
            if is_num:  # remove everything but nums and dots
                try:
                    return float(re.sub(r'[^\d.]+', '', text))
                except ValueError:
                    self.log_error(f"{self.name} - Could not convert value: {text}")
            else:
                return text

    def extract(self, content):
        """Extract data from the content
        """
        raise NotImplementedError

    def save_data(self, data, ignore_check_latest=False):
        self._records.write_records(data,
                                    ignore_check_latest=ignore_check_latest,
                                    context=self.key_context)

    def get_latest(self, uuid):
        return self._records.get_latest(uuid, context=self.key_context)

    def get_records(self, uuid, start, end, filters=None):
        return self._records.get_records(uuid=uuid,
                                         start=start,
                                         end=end,
                                         context=self.key_context)

    def log_debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def log_info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def log_warning(self, msg, *args, **kwargs):
        self.log(logging.WARNING, msg, *args, **kwargs)

    def log_error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def log_critical(self, msg, *args, **kwargs):
        self.log(logging.CRITICAL, msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        _msg = "{} - {}".format(self.name, msg)
        logger.log(level, _msg, *args, **kwargs)


class HTTPSource(BaseSource):

    #: default HTTP method to use when performing the request
    method='GET'
    #: HTTP request timeout value in second
    http_timeout = 10
    #: SSL Verification
    ssl_verify = True


    def __init__(self, name):
        super(HTTPSource, self).__init__(name)

    def fetch(self):
        req = self.get_req()
        res = self.send(req)

        if res is None:
            self._info = None
            self._statuscode = None
            return None

        self._info = res.headers
        self._statuscode = res.status_code

        try:
            # don't use `resp.headers.get('content-length`)
            # because if it doesn't exist in headers it will return None
            # which makes it a false negative
            content_length = res.headers['content-length']
        except KeyError:
            pass
        else:
            if (self._statuscode == 200 and content_length == '0'):
                self.log_warning("Fetched content is empty; skipping cache ...")
                return None

        return self.post_fetch(res)

    def send(self, req, timeout=None):
        """Wrap a :class:`Request <requests.Request>` in a session and send it.
        Any type of error (status code not 2xx or exceptions) will be handled here
        and the function then returns ``None``.

        :param req: :class:`Request <requests.Request>` instance
        :type req: requests.Request
        :param timeout: (optional) How long to wait for the server to send data
            before giving up
        :type timeout: float or tuple
        :rtype: requests.Response
        """
        if req.url is None:
            self.log_warning("no URL provided, abort fetch")
            return None

        if timeout is None:
            timeout = self.http_timeout

        retry_count = 3
        retry_strategy = Retry(
            total=retry_count,
            status_forcelist=[413, 429, 500, 502, 503, 504],
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)
        s = requests.Session()
        s.mount("https://", adapter)
        s.mount("http://", adapter)

        s = requests.Session()
        prepped = s.prepare_request(req)

        try:
            res = s.send(prepped, timeout=timeout, verify=self.ssl_verify)
        except MaxRetryError:
            self.log_error("fetch failed after {} retries".format(retry_count))
            return None
        except requests.Timeout as e:
            self.log_error("fetch error: timeout ({})".format(e.request.url))
            return None
        except requests.RequestException as e:
            self.log_error("fetch error: {}".format(e.request.url))
            return None

        if res.status_code != requests.codes.ok:
            self.log_error("fetch error: status {} ({})".format(
                    res.status_code,
                    req.url,
                )
            )
            self.log_debug(res.text)
            return

        s.close()

        self.log_info("fetch success: {}".format(res.status_code))
        self.log_debug(res.text)

        return res

    def post_fetch(self, response):
        # The Bytes stream is required for the caching operations
        # response.content is a bytes string
        return io.BytesIO(response.content)

    def get_req(self, **kwargs):
        """Return an :class:`requests.Request` instance
        """
        req = requests.Request(
            self.method,
            self.get_url(**kwargs),
            data=self.get_data(**kwargs),
            json=self.get_json(**kwargs),
            params=self.get_params(**kwargs),
            headers=self.get_headers(**kwargs),
        )
        return req

    def get_url(self, **kwargs):
        """Return the target URL, if ``None`` is returned, the scrape will be
        gracefully stopped.
        """
        return self.target

    def get_data(self, **kwargs):
        return {}

    def get_json(self, **kwargs):
        return None

    def get_params(self, **kwargs):
        return {}

    def get_headers(self, **kwargs):
        headers = {
            'User-Agent': get_rnd_item(settings['UA_FILE']),
        }
        return headers

    def get_status_data(self):
        data = {
            'code': self._statuscode
        }
        if self._info is not None and 'Last-Modified' in self._info:
            data['last-modified'] = self._info['Last-Modified']
        return data



# -*- coding: utf-8 -*-

from celery.utils.log import get_task_logger
import requests


HTTP_TIMEOUT = 10

logger = get_task_logger(__name__)


def fetch(req, timeout=None):
    """Wrap a :class:`Request <requests.Request>` in a session and send it.
    Any type of error (status code not 2xx or exceptions) will be handled here
    and the function then returns ``None``.

    :param req: :class:`Request <requests.Request>` instance
    :type req: requests.Request
    :param timeout: (optional) How long to wait for the server to send data
        before giving up
    :type timeout: float or tuple
    :rtype: requests.Response
    """
    if timeout is None:
        timeout = HTTP_TIMEOUT

    s = requests.Session()
    prepped = s.prepare_request(req)
    try:
        res = s.send(prepped, timeout=timeout)
    except requests.Timeout as e:
        logger.error("fetch error: timeout ({})".format(e.request.url))
        return
    except RequestException as e:
        logger.error("fetch error: {}".format(e.msg))
        return

    if res.status_code != requests.codes.ok:
        logger.error("fetch error: status {} ({})".format(
                res.status_code,
                req.url,
            )
        )
        logger.debug(res.text)
        return

    logger.debug("fetch success: {}".format(res.text))

    return res



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from datetime import datetime
from inspect import getmodulename
import logging
import re

from celery.utils.log import get_task_logger

from .base import HTTPSource

import bs4


logger = get_task_logger(__name__)


class Source(HTTPSource):
    """Source class for pm25.in providing China environment data

    Extract data from http://pm25.in/, data table rows are as follow:

    * 0 - Monitoring name
    * 1 - AQI
    * 2 - AQI category in Chinese
    * 3 - primary pollutant
    * 4 - PM2.5
    * 5 - PM10
    * 6 - CO
    * 7 - NO2
    * 8 - O3 (1h average)
    * 9 - O3 (8h average)
    * 10 - SO2
    """

    key_context = {
        u'moduuid': 'pm25in',
    }
    #: regexp matching null characters
    null_re = re.compile(r'_')

    def extract(self, content):
        # Avoid warnings of bs4>=4.4 by explicitly stating parser
        soup = bs4.BeautifulSoup(content, "html5lib")
        time = soup.find(class_="live_data_time").p.text
        ts = datetime.strptime(time.split(u'\uff1a')[1],
                               '%Y-%m-%d %H:%M:%S').replace(tzinfo=self._tz)
        data = {}
        for row in soup.find(id='detail-data').find_all('tr')[1:]:
            cells = row.find_all('td')
            try:
                station = self.get_station_uuid(cells[0].string)
                data[station] = []
            except KeyError as e:
                logger.warning("Station not found ({}); skipping ..."
                               .format(e.message))
            else:
                data[station].append(
                    {
                        'ts': ts,
                        'fields': {
                            'pm25': self.pythonify(cells[4].string, is_num=True),
                            'pm10': self.pythonify(cells[5].string, is_num=True),
                            'co': self.pythonify(cells[6].string, is_num=True),
                            'no2': self.pythonify(cells[7].string, is_num=True),
                            'o3_1h': self.pythonify(cells[8].string, is_num=True),
                            'o3_8h': self.pythonify(cells[9].string, is_num=True),
                            'so2': self.pythonify(cells[10].string, is_num=True),
                        },
                    }
                )
        return data



# -*- coding: utf-8 -*-
import argparse
import distutils.spawn
import os
import sys

from celery.bin.celery import main as clry_main
from celery.utils.imports import find_module, import_from_cwd, NotAPackage

# fix path issues when running `okq-server`
sys.path.insert(0, os.getcwd())


def load_confmod(parser, confmod):
    # import here so we can fix the sys.path when running the script directly
    from openkongqi.conf import config_from_object

    if confmod is None:
        parser.print_help()
        sys.exit(1)

    try:
        find_module(confmod)
    except (ImportError, NotAPackage):
        sys.stderr.write("configuration parameter is not a module"
                "({})\n".format(confmod))
        parser.print_help()
        sys.exit(1)

    mod = import_from_cwd(confmod)
    config_from_object(mod)


def okq_init():
    parser = argparse.ArgumentParser(
        description="initialize database for records",
        epilog="Any extra argument is directly passed to celery, celery's"
        " `--config' argument will be replaced by the value of '--okqconf'")
    parser.add_argument('--okqconf', dest='confmod', action='store',
                        type=str, help='path to a configuration file')
    args = parser.parse_args()

    load_confmod(parser, args.confmod)

    # run magic configuration after environment variable is set
    import openkongqi.conf
    openkongqi.conf.recsdb.db_init()


def okq_server():
    # import here so we can fix the sys.path when running the script directly
    from openkongqi.exceptions import OpenKongqiError

    parser = argparse.ArgumentParser(
        description='outdoor air quality data',
        epilog="Any extra argument is directly passed to celery, celery's"
        " `--config' argument will be replaced by the value of '--okqconf'")
    parser.add_argument('--okqconf', dest='confmod', action='store',
            type=str, help='path to a configuration module (mandatory)')
    args, clry_args = parser.parse_known_args()

    load_confmod(parser, args.confmod)

    # reproduce sys.argv to call celery
    clry_args.insert(0, distutils.spawn.find_executable('celery'))
    try:
        clry_main(clry_args)
    except OpenKongqiError as e:
        sys.stderr.write(e.message + "\n")
        sys.exit(1)


if __name__ == '__main__':          # pragma: no cover
    okq_server()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from .sqlalch import SQLAlchemyRecordsWrapper

from sqlalchemy.engine.url import URL

_NAME = 'openkongqi'


class RecordsWrapper(SQLAlchemyRecordsWrapper):

    def create_dsn(self, settings):
        engine = 'sqlite'
        db_name = settings.get('NAME', _NAME)
        db_filename = '{}.db'.format(db_name)
        if db_name == ':memory:':
            dsn = URL(engine)  # in memory sqlite db
        else:
            dsn = URL(engine, database=db_filename)
        return dsn



# -*- coding: utf-8 -*-

from ..conf import settings


def get_source(name):
    """Return a single source information.

    :param name: source name as used in the configuration
    :type name: str
    :returns: info - a dict of information about this source
    """
    info = settings['SOURCES'][name]
    info['name'] = name
    return info


def get_sources():
    """Return a generator going over the sources and returning source info.

    :returns: generator - sources information
    """
    for name in settings['SOURCES'].keys():
        yield get_source(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Command line tool to test whether extraction method works.

The user is to run the tests and compare the web data with the neatly
display data visually.

The table format will always appear as TIME, STATION, + accompanying
particular measurements (depends on the source).
"""

from __future__ import absolute_import, print_function, unicode_literals
from datetime import datetime
import argparse
import os

from openkongqi.utils import get_source

HTTP_TIMEOUT = 30

SOURCES = None


def fetch_and_extract(src_name):
    src = get_source(src_name)
    src_content = src.fetch()
    data = src.extract(src_content)
    return data


def get_categories(data):
    categories_set = set(
        reduce(lambda x, y: x + y,
               [r['fields'].keys()
                for uuid, records in data.items()
                for r in records])
    )
    return sorted(categories_set)


def display_table(data):
    # get info on categories for correct row format
    categories = get_categories(data)
    row_fmt = "{:^25} " + "{:^55} " + "{:>8} " * len(categories)
    col_names = ("TIME", "STATION", ) + tuple([c.upper() for c in categories])
    table = "\n" + row_fmt.format(*col_names)
    table += "\n" + row_fmt.format(*tuple(["-" * len(cn) for cn in col_names]))
    for uuid, records in data.items():
        for record in records:
            time = datetime.strftime(record['ts'], "%Y-%m-%d %H:%M:%S")
            row_data = (time, uuid, ) + tuple(record['fields'].get(c, '')
                                              for c in categories)
            table += "\n" + row_fmt.format(*row_data)
    print(table)


def create_parser():
    """Return command-line parser."""
    # setup parser
    parser = argparse.ArgumentParser()
    # list available sources
    parser.add_argument("-l", "--list-sources",
                        action="store_true",
                        help="list sources that are available")
    # okq custom configuration file
    parser.add_argument('--okqconf', dest='conffile', action='store',
                        type=str, help='path to a configuration file')
    # main input
    parser.add_argument("sources", nargs="*",
                        help="Name of a valid source. "
                        "See `okqconfig.py` for more help.",
                        type=str)

    return parser


def parse_args():
    """Parse command-line options."""
    parser = create_parser()
    args = parser.parse_args()

    if not args.sources and not args.list_sources:
        parser.error("Incorrect number of arguments.")

    if args.conffile is not None:
        os.environ.setdefault('OKQ_CONFMODULE', args.conffile)

    # run magic config after envvar is set
    from openkongqi.conf import settings
    global SOURCES
    SOURCES = settings['SOURCES']

    # print available choices, and exit program
    if args.list_sources:
        print()
        print("LIST OF AVAILABLE SOURCES")
        print("=========================")
        for src in sorted(SOURCES.keys()):
            print("    *  {}".format(src))
        print()
        exit()

    return args


def main():
    """Command-line entry."""
    args = parse_args()

    print()

    # ==== START :: Check if user selected all ====
    #
    # i.e.) "pm25.in:*" searches all keys beginning with "pm25.in:"
    #       if exists, then expand these selectors with their "children"
    #
    # ['aqhi.gov.hk:central', 'pm25.in:*] ==>
    # ['aqhi.gov.hk:central',
    #     'pm25.in:bejing', 'pm25.in:chengdu',
    #     'pm25.in:shanghai', 'pm25.in:yingkou']
    i = len(args.sources) - 1
    while i >= 0:
        if args.sources[i][-1] == "*":
            common_header = args.sources[i][:-1]
            children_sources = filter(
                lambda s: s[:len(common_header)] == common_header,
                SOURCES.keys()
            )
            args.sources = args.sources[:i] + \
                children_sources + args.sources[i + 1:]
            i += len(children_sources)
        i -= 1
    # ==== END :: Check if user selected all ====

    print("Selecting sources in order: {}".format(sorted(args.sources)))

    # fetch, extract, and display data
    for source in sorted(args.sources):
        data = fetch_and_extract(source)
        display_table(data)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

from .conf import settings

_API_KEYS = settings['API_KEYS']


def get_api_key(provider=None):
    """Get the api key map, given a provider

    :param provider: a provider for the API Key
    :type provider: str|unicode
    """
    try:
        api_key = _API_KEYS[provider]
    except KeyError:
        return None
    return api_key



# -*- coding: utf-8 -*-


class OpenKongqiError(Exception):
    pass


class ConfigError(OpenKongqiError):
    pass


class CacheError(OpenKongqiError):
    pass


class SourceError(OpenKongqiError):
    pass


class FeedError(OpenKongqiError):
    pass


class APIKeyNotFoundError(OpenKongqiError):
    pass
